
   import external.Viewer;
   import external.viewer.*;
   import java.awt.GridBagConstraints;
   import java.awt.GridBagLayout;
   import java.awt.Insets;
   import javax.swing.BorderFactory;
   import javax.swing.JComponent;
   import javax.swing.JLabel;
   import javax.swing.JPanel;
   import javax.swing.JScrollPane;
   import javax.swing.JTextField;

    public abstract class FloatView extends JScrollPane
            implements Viewer {
   
      private JLabel top_;
      private JLabel[] textSegments_ = new JLabel[12];
      private JLabel[] bottom_ = new JLabel[7];
   
      private int signBit_, firstExponentBit_, lastExponentBit_;
      private int firstMantissaBit_, lastMantissaBit_;
      private int exponentBias_;
   
       public FloatView(int sign_bit, int first_exponent_bit,
            int last_exponent_bit, int first_mantissa_bit,
            int last_mantissa_bit, int exponent_bias) {
         super();
         
         signBit_ = sign_bit;
         firstExponentBit_ = first_exponent_bit;
         lastExponentBit_ = last_exponent_bit;
         firstMantissaBit_ = first_mantissa_bit;
         lastMantissaBit_ = last_mantissa_bit;
         exponentBias_ = exponent_bias;
         
         JPanel content = new JPanel(false);
         GridBagLayout layout = new GridBagLayout();
         content.setLayout(layout);
         GridBagConstraints constraints = new GridBagConstraints();
         Insets insets = constraints.insets;
        
         int spacing = 8;
         constraints.weightx = constraints.weighty = .001;
         insets.bottom = 0;
         insets.top = spacing;
         insets.left = insets.right = spacing;
         constraints.gridwidth = GridBagConstraints.REMAINDER;
         constraints.fill = GridBagConstraints.HORIZONTAL;
         constraints.anchor = GridBagConstraints.WEST;
      
         top_ = new JLabel(" ");
         content.add(top_);
         layout.setConstraints(top_, constraints);
      
         constraints.gridwidth = 1;
         int seg = 0;
         for(int row = 0; row < 4; row++) {
            if(row == 1)
               insets.top = 0;
            for(int col = 0; col < 3; col++) {
               JLabel segment = new JLabel(" ");
               segment.setBorder(BorderFactory.createEtchedBorder());
               if(row >= 1)
                  segment.setHorizontalAlignment(JTextField.RIGHT);
               textSegments_[seg++] = segment;
               insets.left = (col == 0)?spacing:0;
               insets.right = (col == 2)?spacing:0;
               content.add(segment);
               layout.setConstraints(segment, constraints);
               if(col == 2) {
                  JPanel spacer = new JPanel(false);
                  constraints.gridwidth = GridBagConstraints.REMAINDER;
                  constraints.weightx = 1.0;
                  content.add(spacer);
                  layout.setConstraints(spacer, constraints);
                  constraints.gridwidth = 1;
                  constraints.weightx = .001;
               }
            }
         }
         textSegments_[0].setText(" Sign");
         textSegments_[1].setText(" Exponent");
         textSegments_[2].setText(" Mantissa");
      
         insets.left = insets.right = insets.top = spacing;
         constraints.gridwidth = GridBagConstraints.REMAINDER;
         for(int row = 0; row < 7; row++) {
            if(row == 1)
               insets.top = 0;
            bottom_[row] = new JLabel(" ");
            content.add(bottom_[row]);
            layout.setConstraints(bottom_[row], constraints); }
         constraints.weighty = 1.0;
         constraints.fill = GridBagConstraints.BOTH;
         JPanel spacer = new JPanel(false);
         content.add(spacer);
         layout.setConstraints(spacer, constraints);
         
         setViewportView(content);
      }
       
       public JComponent getComponent() {
         return this; }
          
       public String getName() {
         return "Detail"; }
   
       public int getPriority() {
         return -1; }
   
       public void setFrozen() {}
   
       public void update(Value value, DebugContext context) throws ViewerException {
         String value_text = getValueText(value, context);
         top_.setText("value = " + value_text);
         long bits = getBits(value, context);
         
         long sign_mask = 1l << signBit_;
         bottom_[0].setText("sign = " + (((bits & sign_mask) == 0l)?'+':'-'));
         
         textSegments_[3].setText(((bits & sign_mask) == 0l)?" 0":" 1");
         textSegments_[4].setText(" " + bitString(bits, firstExponentBit_, lastExponentBit_));
         textSegments_[5].setText(" " + bitString(bits, firstMantissaBit_, lastMantissaBit_));
      
         long exponent_mask = (2l << firstExponentBit_) - (1l << lastExponentBit_);
         long exponent = (bits & exponent_mask) >> lastExponentBit_;
         bottom_[1].setText("exponent = " + exponent + " - bias of " + exponentBias_ + " = " + (exponent - exponentBias_));
            
         long mantissa_mask = (2l << firstMantissaBit_) - (1l << lastMantissaBit_);
         long mantissa = (bits & mantissa_mask) >> lastMantissaBit_;
         double mant_double = (double)mantissa / (2l << (firstMantissaBit_ - lastMantissaBit_));
         
         textSegments_[6].setText(((bits & sign_mask) == 0l)?" 0":" 1");
         textSegments_[7].setText("0x" + Long.toString(exponent, 16));
         textSegments_[8].setText("0x" + Long.toString(mantissa, 16));
         
         textSegments_[9].setText(((bits & sign_mask) == 0l)?" 0":" 1");
         textSegments_[10].setText(Long.toString(exponent));
         textSegments_[11].setText(Long.toString(mantissa));
         
         bottom_[2].setText("mantissa = assumed 1 + " + mantissa + " / 2^" + (firstMantissaBit_ - lastMantissaBit_));
         bottom_[3].setText("      = (approximately) " + getMantissaText(mant_double + 1));
         bottom_[4].setText("value = (sign) mantissa * 2 ^ exponent");
         bottom_[5].setText("      = " + (((bits & sign_mask) == 0l)?'+':'-') +
                  getMantissaText(mant_double + 1) + " * 2 ^ " + (exponent - exponentBias_));
         bottom_[6].setText("      = " + value_text);
      }
   
   
       private String bitString(long bits, int first_bit, int last_bit) {
         StringBuffer buffer = new StringBuffer();
         long bit = 1l << first_bit;
         for(int i = first_bit; i >= last_bit; i--, bit >>= 1) {
            buffer.append(((bits & bit) == 0l)?'0':'1');
            if(i > last_bit && ((i - last_bit) % 4) == 0)
               buffer.append(' ');
         }
         return buffer.toString();
      }
      
      
       public abstract String getValueText(Value value, DebugContext context)
            throws ViewerException;
      
       public abstract long getBits(Value value, DebugContext context)
            throws ViewerException;
       
       public abstract String getMantissaText(double mantissa);
   }
